/* Etherboot routines for PCBIOS firmware.
 *
 * Body of routines taken from old pcbios.S
 */

#ifdef PCBIOS

#include "etherboot.h"
#include "realmode.h"
#include "segoff.h"


char bios_version[30] = "EthrLink ver2.0 - 30.11.2009";
#define CF ( 1 << 0 )

/**************************************************************************
CURRTICKS - Get Time
Use direct memory access to BIOS variables, longword 0040:006C (ticks
today) and byte 0040:0070 (midnight crossover flag) instead of calling
timeofday BIOS interrupt.
**************************************************************************/
#if defined(CONFIG_TSC_CURRTICKS)
#undef CONFIG_BIOS_CURRTICKS
#else
#define CONFIG_BIOS_CURRTICKS 1
#endif
#if defined(CONFIG_BIOS_CURRTICKS)
unsigned long currticks (void) 
{
	static uint32_t days = 0;
	uint32_t *ticks = VIRTUAL(0x0040,0x006c);
	uint8_t *midnight = VIRTUAL(0x0040,0x0070);

	/* Re-enable interrupts so that the timer interrupt can occur
	 */
	RM_FRAGMENT(rm_currticks,
		"sti\n\t"
		"nop\n\t"
		"nop\n\t"
		"cli\n\t"
	);	

	real_call ( rm_currticks, NULL, NULL );

	if ( *midnight ) {
		*midnight = 0;
		days += 0x1800b0;
	}
	return ( days + *ticks );
}
#endif	/* CONFIG_BIOS_CURRTICKS */


/**************************************************************************
GET_GEOMETRY - Read a NAND Flash disk geometry
**************************************************************************/
unsigned char nand_get_geometry ( int drive, int *cylinder, int *head, int *sector)
{
	unsigned long tmp;
	struct {
		reg16_t ax;
		reg16_t cx;
		reg16_t dx;
		segoff_t buffer;
	} PACKED in_stack;
	struct {
		reg16_t bx;
		reg16_t cx;
		reg16_t dx;
		segoff_t buffer;
		reg16_t flags;
	} PACKED out_stack;
	reg16_t ret_ax;

	RM_FRAGMENT(rm_pcbios_get_geometrty,
		"sti\n\t"
		"popw %ax\n\t"
		"popw %cx\n\t"
		"popw %dx\n\t"
		"popw %bx\n\t"
		"popw %es\n\t"
		"int $0x13\n\t"
		"pushfw\n\t"
		"pushw %di\n\t"
		"pushw %es\n\t"
		"pushw %dx\n\t"
		"pushw %cx\n\t"
		"pushw %bx\n\t"
		"cli\n\t"
	);

	in_stack.ax.h = 0x08; /* INT 13,8 - Read disk parameters */
	in_stack.dx.l = drive;

	ret_ax.word = real_call ( rm_pcbios_get_geometrty, &in_stack, &out_stack );
	
	*cylinder = out_stack.cx.h + ((out_stack.cx.l & 0xc0) << 2) + 1;
    *head = out_stack.dx.h  + 1;
    *sector = out_stack.cx.l & 0x3f;

/*
	//printf ("Nand geometry parameters before: C/H/S = %d/%d/%d.\n", *cylinder, *head,*sector);
	tmp = (*cylinder)*(*head)*(*sector);
	//printf ("tmp=%x.\n", tmp);

	*cylinder = ((tmp/255)/63);
	//printf ("cylinder=%x.\n", *cylinder);

	*head = 255;//((tmp /(*cylinder))/63);
	*sector = 63;//((tmp /(*cylinder))/(*head));
	
*/
	return ( out_stack.flags.word & CF ) ? ret_ax.word : 0;
}


/**************************************************************************
FLASH_FIND - Detect NAND Flash disk number
**************************************************************************/
unsigned char nand_flash_find (void)
{
	int i;
	unsigned char *drive_detect_addr;



#ifdef CM_IAM
int timeout = 10;
int ready =0;
  char port;
  char buffer[512];

   char firmware[9];
   char model[41];
    unsigned short read_val,c,h,s;
    unsigned long total, devlba,total_calc;
     unsigned int devsize;
     int c1,h1,s1;
    int x;

    printf("Detecting NAND...\n");
    outb(0xA0,0x1f6);
  
    while(timeout)
    {
    port = inb(0x3f6);
      if(port==((char)0x50))
          {
            
            ready =1;
            break;
          }
      timeout--;
      sleep(1);
    
    }
    if(ready==0)
    {
      printf("NAND drive isn't ready on primary master...skip\n");
      return 0xff;
    
    }
    
    outb(0xEC,0x1F7);//identify
    timeout =10;
    ready =0;
    while(timeout)
    {
    port = inb(0x1f7);
    // printf("%x %x %x  \n\r",(int)port,(char)(port&0x80),(char)(port&0x8));
      if(((port&0x80)==0)&& ((port&0x8)==8))
          {
            
            ready =1;
            break;
          }
      timeout--;
      sleep(1);
    
    }
      if(ready==0)
    {
       printf("NAND drive isn't ready on primary master...skip\n");
        return 0xff;
    
    }
    
      for (x=0;x<512;x++)
    {
      buffer[x] =0;
    }
    //read buffer
    printf("Get ID...");
    for (x=0;x<256;x+=2)
    {
       read_val=inw(0x1f0);

      
       
      buffer[x+1] =(char) (read_val&0xff);
      buffer[x] = (char)((read_val>>8)&0xff);
      
     }
      printf("OK\n");
   

    firmware[8]='\0';
    model[40]='\0';
     
     strncpy(firmware,&buffer[46],8);
     strncpy(model,&buffer[54],40);

     c= buffer[0x3]&0xff|(buffer[0x2]&0xffff)<<8;
     h= buffer[0x7]&0xff|(buffer[0x6]&0xffff)<<8;
      s= buffer[0xd]&0xff|(buffer[0xc]&0xffff)<<8;
    total_calc = ((unsigned long)c*(unsigned long)h*(unsigned long)s)>>20;
    
    if(total_calc > 0x7)
    {
      printf("Nand flash isn't present / disabled in the BIOS...%lx chs= %x %x %x skip\n",total_calc,c,h,s);
      return 0xFF;
    }
    if(total_calc<0x2)
    {
       elink.auto_config.enable_split =0;
    }
    
   
      devlba = *((unsigned long *)&buffer[0x78]);  //Num of LBA sectors
      total = devlba;
      devlba = devlba>>29 ;
      devsize =(unsigned int)devlba +1;

    printf ("Firmware revision : %s\n",firmware);
    printf ("Model: %s\n",model);
    printf ("NAND size: %dGb (%lx bytes)   \n",(int)devsize,total*2 );
    
 
   //looking for the same  size on the int13
	
   for(x=0x80;x<0x85;x++)
  {

	nand_get_geometry (x, &c1, &h1, &s1);
 
	total=((unsigned long)c1*(unsigned long)h1*(unsigned long)s1)>>20;
	//printf("disk %x chs = %x %x %x total = %lx calc =%lx\n",x,c1,h1,s1,total,total_calc);
	
  if(total==total_calc)
    {
    printf("Found on 0x%02x \n",x);
    return x;
    }
  }

  return 0xff;
//return 0x81;

#endif
}

/**************************************************************************
FLASH_IOCTL - Sends Flash control commands
**************************************************************************/
unsigned char nand_flash_ioctl (int drive, int cmd)
{
	struct {
		reg16_t ax;
		reg16_t cx;
		reg16_t dx;
	} PACKED in_stack;
	struct {
		reg16_t flags;
	} PACKED out_stack;
	reg16_t ret_ax;

	RM_FRAGMENT(rm_pcbios_flash_ioctl,
		"sti\n\t"
		"popw %ax\n\t"
		"popw %cx\n\t"
		"popw %dx\n\t"
		"int $0x13\n\t"
		"pushfw\n\t"
		"cli\n\t"
	);


	
	return 1;
}

/**************************************************************************
FLASH_READ - Read NAND Flash sector
**************************************************************************/

unsigned char nand_flash_read ( int drive, int cylinder, int head, int sector,int sector_num,char *buf ) 
{
	struct {
		reg16_t ax;
		reg16_t cx;
		reg16_t dx;
		segoff_t buffer;
	} PACKED in_stack;
	struct {
		reg16_t flags;
	} PACKED out_stack;
	reg16_t ret_ax;

	RM_FRAGMENT(rm_pcbios_flash_read,
		"sti\n\t"
		"popw %ax\n\t"
		"popw %cx\n\t"
		"popw %dx\n\t"
		"popw %bx\n\t"
		"popw %es\n\t"
		"int $0x13\n\t"
		"pushfw\n\t"
		"cli\n\t"
	);

	led_ioctl(TOGGLE,2048);

	in_stack.ax.h = 2; /* INT 13,2 - Read disk sector */
	in_stack.ax.l = sector_num; /* Read one sector */
	in_stack.cx.h = (char)(cylinder & 0xff);//cylinder & 0xff;
	in_stack.cx.l = (char)(((cylinder & 0x300) >> 2) | (sector & 0x3f));//( ( cylinder >> 8 ) & 0x3 ) | sector;
	in_stack.dx.h = head;
	in_stack.dx.l = drive;
	in_stack.buffer.segment = SEGMENT ( buf );
	in_stack.buffer.offset = OFFSET ( buf );

	ret_ax.word = real_call ( rm_pcbios_flash_read, &in_stack, &out_stack );
	/*
	int i=0;
	for (i=0;i<512;i++)
		buf[i]=65;
	for (i=0;i<512;i++)
		buf[i+512]=66;
	printf ("Segment=%x, offset=%x (buf[0]=%d,buf[1]=%d).\n",SEGMENT ( buf ),OFFSET ( buf ),buf[0],buf[1]);
	*/
	return ( out_stack.flags.word & CF ) ? ret_ax.word : 0;
}


/**************************************************************************
FLASH_WRITE - Write NAND Flash sector
**************************************************************************/

unsigned char nand_flash_write ( int drive, int cylinder, int head, int sector,int sector_num,char *buf ) 
{
	struct {
		reg16_t ax;
		reg16_t cx;
		reg16_t dx;
		segoff_t buffer;
	} PACKED in_stack;
	struct {
		reg16_t flags;
	} PACKED out_stack;
	reg16_t ret_ax;

	RM_FRAGMENT(rm_pcbios_flash_write,
		"sti\n\t"
		"popw %ax\n\t"
		"popw %cx\n\t"
		"popw %dx\n\t"
		"popw %bx\n\t"
		"popw %es\n\t"
		"int $0x13\n\t"
		"pushfw\n\t"
		"cli\n\t"
	);

	//return 1 ;
	led_ioctl(TOGGLE,2048);

	in_stack.ax.h = 3; /* INT 13,2 - Write disk sector */
	in_stack.ax.l = sector_num; /* Read one sector */
	in_stack.cx.h = (char)(cylinder & 0xff);//cylinder & 0xff;
	in_stack.cx.l = (char)(((cylinder & 0x300) >> 2) | (sector & 0x3f));//( ( cylinder >> 8 ) & 0x3 ) | sector;
	in_stack.dx.h = head;
	in_stack.dx.l = drive;
	in_stack.buffer.segment = SEGMENT ( buf );
	in_stack.buffer.offset = OFFSET ( buf );

	//printf ("Segment=%x, offset=%x.\n",SEGMENT ( buf ),OFFSET ( buf ));
	ret_ax.word = real_call ( rm_pcbios_flash_write, &in_stack, &out_stack );
	return ( out_stack.flags.word & CF ) ? ret_ax.word : 0;
}

/**************************************************************************
NOR_FLASH_IOCTL - Perform NOR Flash block IO
**************************************************************************/
int setbyte(unsigned int , unsigned char );
int erasesec(unsigned long );
int eraseblock(unsigned long );
unsigned char nor_flash_ioctl ( int function, unsigned int address,unsigned int count,char *buf) 
{


	if (function <=0 || function >=4 )
		return 0;
		led_ioctl(TOGGLE,0);
		//outl(0x1210,2);

	struct {
		reg16_t ax;
		reg16_t cx;
		segoff_t buffer;
		segoff_t address;
	} PACKED in_stack;
	struct {
		reg16_t flags;
	} PACKED out_stack;
	reg16_t ret_ax;

	
	#ifdef CM_IAM
	RM_FRAGMENT(rm_pcbios_disable_int,
		"cli\n\t"
	);
	RM_FRAGMENT(rm_pcbios_enable_int,
		"sti\n\t"
	);
	
	unsigned int loc_count=0;
	unsigned char tmp;
	
  

real_call ( rm_pcbios_disable_int,&in_stack, &out_stack);
		
	switch(function)
	{
    case 0:
      {
      
      }
      break;
    case 1:
      {//erase function
      
  
     
        if(count <0x10000 )
          {
            for(loc_count=0;loc_count<count;loc_count+=0x1000)
            {
              erasesec( address+loc_count);
            }
          }
          else
            eraseblock( address);
            
           // 	outl(0x1210,0x18);
           real_call(rm_pcbios_enable_int,&in_stack, &out_stack);
          return 1;
        
      }
      break;
    case 2:
      {//read function
       memcpy (buf,phys_to_virt(address),count);
         real_call(rm_pcbios_enable_int,&in_stack, &out_stack);
 
       return 1;
      }
      break;
    case 3:
      {//write function
      
      //return 1;
         //   printf("Writing block %lx size = %x\n",address,count);
         printf(".");
          outb(0x55,0x80);
            for(loc_count=0; loc_count<count; loc_count++)
             {
          
               // setbyte((unsigned long)(address+loc_count), *((unsigned char *)(buf+loc_count)));
                setbyte(address+loc_count, (unsigned char)buf[loc_count]);
                  
              }
         printf("\b");
      real_call(rm_pcbios_enable_int,&in_stack, &out_stack);
  
        return 1;
      }
      break;
      
      
    default: break;
  
  
   
  
  }
	#endif
	
	
}
/*---------------------------------------------------------*/
#define byte unsigned char


int erasesec(unsigned long addr)
{
    char i;
    unsigned long  lockreg, lockreg2, eraseaddr;
    int x;
    eraseaddr =addr& 0xFFFF0000;
   // printf("Erasing sector %lx...", addr);



	// Unlocking software lock registers

	  lockreg = (addr & 0xff0f0000) | 0x00b00002;
	 // lockreg2 = lockreg | 0x00008000;
 

    *((byte *)phys_to_virt(lockreg)) = 0x00;
    *((byte *)phys_to_virt(lockreg2)) = 0x00;    
     
   
    
     i= *((byte*)phys_to_virt(addr+1));
     
     *((byte *)phys_to_virt(eraseaddr|0x00005555))=0xAA;// Erase, first cycle
      *((byte *)phys_to_virt(eraseaddr|0x00002AAA))=0x55;// Erase, second cycle
      *((byte *)phys_to_virt(eraseaddr|0x00005555))=0x80;// Erase, third cycle 
      *((byte *)phys_to_virt(eraseaddr|0x00005555))=0xAA;// Erase, 4th cycle
      *((byte *)phys_to_virt(eraseaddr|0x00002AAA))=0x55;// Erase, 5th cycle  
      *((byte *)phys_to_virt(addr))=0x30;// Sector erase address      
   
          
                             
    
//		printf(".processing");
		for ( x=0;x<10;x++)
		{
   
//		printf(".");
		}
     sleep(1);
  
       	while(*((byte *)phys_to_virt(addr))!=0xFF)
       {   
        
   //      printf("addr =%lx value %x\n",addr,*((byte *)phys_to_virt(addr)) );
      for ( x=0;x<50000;x++);
       };
	//	printf("Erase complete.\n");

    return 0;
}
/*---------------------------------------------------------*/
int eraseblock(unsigned long addr)
{
    unsigned  char i;
    unsigned long  lockreg, lockreg2,eraseaddr;
    int x;

    eraseaddr =addr& 0xFFFF0000;
  //  printf("Erasing block %lx...", addr);

	// Unlocking software lock registers

	  lockreg = (addr & 0xff0f0000) | 0x00b00002;
	  lockreg2 = lockreg | 0x00008000;
 
    *((byte *)phys_to_virt(lockreg)) = 0x00;
    *((byte *)phys_to_virt(lockreg2)) = 0x00; 
       
       i= *((byte*)phys_to_virt(addr));
       i= *((byte*)phys_to_virt(addr));

      *((byte *)phys_to_virt(eraseaddr|0x00005555))=0xAA;// Erase, first cycle
      *((byte *)phys_to_virt(eraseaddr|0x00002AAA))=0x55;// Erase, second cycle
      *((byte *)phys_to_virt(eraseaddr|0x00005555))=0x80;// Erase, third cycle 
      *((byte *)phys_to_virt(eraseaddr|0x00005555))=0xAA;// Erase, 4th cycle
      *((byte *)phys_to_virt(eraseaddr|0x00002AAA))=0x55;// Erase, 5th cycle  
      *((byte *)phys_to_virt(addr))=0x50;// Sector erase address
                             
    
//		printf(".processing");
		for ( x=0;x<10;x++)
		{
	//	printf(".");
		}
	 sleep(1);
    	while(*((byte *)phys_to_virt(addr))!=0xFF)
       {
       for ( x=0;x<50000;x++);
     //  sleep(1);
   
    //    printf("value %x\n",*((byte *)phys_to_virt(addr)) );
       };
//		printf("Erase complete.\n");

    return 0;
}

int setbyte(unsigned int addr, unsigned char data)
{
    unsigned char ch,i;
    unsigned int xx=0,yy=0;
    unsigned int eraseaddr;
    eraseaddr =addr& 0xFFFF0000;
    int timeout;

    ch = *((byte *)phys_to_virt(addr));
    if((data & ch) != data)
    {
	   	printf("Media not empty: addr: %lx data: %x\n",addr,ch);
		  return 0;
    }
    
    
	
      i= *((byte*)phys_to_virt(addr));
    
      *((byte *)phys_to_virt(eraseaddr|0x00005555))=0xAA;
      *((byte *)phys_to_virt(eraseaddr|0x00002AAA))=0x55;
      *((byte *)phys_to_virt(eraseaddr|0x00005555))=0xA0;  
      *((byte *)phys_to_virt(addr))=data;       
           
           
         	for(xx=0;xx<50000;xx++) 
           for(yy=0;yy<10;yy++)  ;
         	
       i= *((byte*)phys_to_virt(addr));
       timeout = 100;
	    	while( i != data )
      {		
	 	  	for(xx=0;xx<50000;xx++) 
	 	    // 	for(yy=0;yy<100;yy++); //delay 

         i= *((byte*)phys_to_virt(addr));
         outb(i,0x80);
         if(timeout<1)
         {
               printf("timeout: addr = %lx data = %x real data =%x\n",addr, data,*((byte *)phys_to_virt(addr)));
               timeout=100;
         }
         timeout--;
          //    printf("addr = %lx data = %x real data =%x\n",addr, data,*((byte *)phys_to_virt(addr)));
			      // 	for( xx=0;xx<60000;xx++);	//	printf("Verifying data: addr: %lx data: %x\r",addr,ch);
	    }
	//    printf("addr = %lx data = %x real data =%x\n",addr, data,*((byte *)phys_to_virt(addr)));
		//	printf("Verifying data: readed: addr %lx %lx must be: %lx\n",ExtMem_ReadByte(addr),(int)data);
	

    return 1;
}



/**************************************************************************
NOR_FLASH_IOCTL - Perform NOR Flash block IO
**************************************************************************/

unsigned char nor_flash_unlock () 
{


#ifdef CM_IAM
 /*outl(0xCF8,0x8000f8d8);
 printf ("inital value =%lx " ,inl(0xCFC));
 outl(0xCFC,inl(0xCFC)|0x00000001);
  printf ("final value =%lx \n" ,inl(0xCFC));*/

  // printf ("MFGID =%x , DEVID = %x\n",mfgid, devid);
   
#endif

	return 1;
}

/**************************************************************************
NOR_FLASH_IOCTL - Perform NOR Flash block IO
**************************************************************************/

void  bios_clear_screen () 
{
	int i;
	

	struct {
		reg16_t ax;
		reg16_t bx;
		reg16_t cx;
		reg16_t dx;
	} PACKED in_stack;
	struct {
		reg16_t flags;
	} PACKED out_stack;
	reg16_t ret_ax;

	RM_FRAGMENT(rm_pcbios_set_cursor,
		"sti\n\t"
		"popw %ax\n\t"
		"popw %bx\n\t"
		"popw %cx\n\t"
		"popw %dx\n\t"
		"int $0x10\n\t"
		"pushfw\n\t"
		"cli\n\t"
	);

	RM_FRAGMENT(rm_pcbios_clear_screen,
		"sti\n\t"
		"popw %ax\n\t"
		"popw %bx\n\t"
		"popw %cx\n\t"
		"popw %dx\n\t"
		"int $0x10\n\t"
		"pushfw\n\t"
		"cli\n\t"
	);
/*
	for (i=0;i<25;i++)
		putchar('\n');
*/
	in_stack.ax.h = 0x06; /* INT 10,6 - Scroll up*/
	in_stack.ax.l = 0x00; 
	in_stack.bx.h = 0x07; 
	in_stack.cx.h = 0x00; 
	in_stack.cx.l = 0x00; 
	in_stack.dx.h = 0x19; 
	in_stack.dx.l = 0x50; 
	
	ret_ax.word = real_call ( rm_pcbios_clear_screen, &in_stack, &out_stack );

	in_stack.ax.h = 0x02; /* INT 10,2 - Set cursor*/
	in_stack.bx.h = 0x00; 
	in_stack.dx.h = 0x00; 
	in_stack.dx.l = 0x00; 
	ret_ax.word = real_call ( rm_pcbios_set_cursor, &in_stack, &out_stack );
	return ;

}


 
 #ifdef CM_IAM
 
 //enum led_state { TOGGLE =1 ,LED_ON=2,LED_OFF=3};
 void led_ioctl (led_state function,int delay_count) 
{

	static int count1 =0;
	static int state1 =0;
  switch(function)
  {
    case TOGGLE:
    {
    		if (!delay_count)
		      	count1 = 0;
	     	if (count1 == 0)
        {
			       count1 = delay_count;
			     if (state1)
            {
			      outb(0x2,0x1189);
				      state1 =0;
		      	}
            else 
            {
			         outb(0x3,0x1189);
			         	state1 =1;
			       }
    
          }
        else
			     count1--;
    
    }break;
    case LED_ON:
    {
     outb(0x2,0x1189);
       
    }break;
      case LED_OFF:
    {
    
    outb(0x3,0x1189);
    }break;
    
    default: break; 
      
  }

  return;
}
 
 #endif



/**************************************************************************
Perform NOR chipset reset 
**************************************************************************/

void  bios_reset_cm_iam () 
{
	

	struct {
		reg16_t ax;
		reg16_t dx;
	} PACKED in_stack;
	struct {
		reg16_t flags;
	} PACKED out_stack;
	reg16_t ret_ax;

	RM_FRAGMENT(rm_pcbios_reset_iam,
		"sti\n\t"
		"popw %ax\n\t"
		"popw %dx\n\t"
		"outb %al,%dx\n\t"
		"pushfw\n\t"
		"cli\n\t"
	);

	
	in_stack.dx.word = 0xcf9; 
	in_stack.ax.word = 0x02; 
	ret_ax.word = real_call ( rm_pcbios_reset_iam, &in_stack, &out_stack );
	return ;

}
/**************************************************************************
INT15 - Call Interrupt 0x15
**************************************************************************/
int int15 ( int ax )
{
	struct {
		reg16_t ax;
	} PACKED in_stack;
	struct {
		reg16_t flags;
	} PACKED out_stack;
	reg16_t ret_ax;

	RM_FRAGMENT(rm_int15,
		"sti\n\t"
		"popw %ax\n\t"
		"stc\n\t"
		"int $0x15\n\t"
		"pushf\n\t"
		"cli\n\t"
	);

	in_stack.ax.word = ax;
	ret_ax.word = real_call ( rm_int15, &in_stack, &out_stack );

	/* Carry flag clear indicates function not supported */
	if ( ! ( out_stack.flags.word & CF ) ) return 0;
	return ret_ax.h;
}

#ifdef	POWERSAVE
/**************************************************************************
CPU_NAP - Save power by halting the CPU until the next interrupt
**************************************************************************/
void cpu_nap ( void )
{
	RM_FRAGMENT(rm_cpu_nap,
		"sti\n\t"
		"hlt\n\t"
		"cli\n\t"
	);	

	real_call ( rm_cpu_nap, NULL, NULL );
}
#endif	/* POWERSAVE */

#if	(TRY_FLOPPY_FIRST > 0)
/**************************************************************************
DISK_INIT - Initialize the disk system
**************************************************************************/
void disk_init ( void )
{
	RM_FRAGMENT(rm_disk_init,
		"sti\n\t"
		"xorw %ax,%ax\n\t"
		"movb $0x80,%dl\n\t"
		"int $0x13\n\t"
		"cli\n\t"
	);
	
	real_call ( rm_disk_init, NULL, NULL );
}

/**************************************************************************
DISK_READ - Read a sector from disk
**************************************************************************/
unsigned int pcbios_disk_read ( int drive, int cylinder, int head, int sector,
				char *buf ) {
	struct {
		reg16_t ax;
		reg16_t cx;
		reg16_t dx;
		segoff_t buffer;
	} PACKED in_stack;
	struct {
		reg16_t flags;
	} PACKED out_stack;
	reg16_t ret_ax;

	RM_FRAGMENT(rm_pcbios_disk_read,
		"sti\n\t"
		"popw %ax\n\t"
		"popw %cx\n\t"
		"popw %dx\n\t"
		"popw %bx\n\t"
		"popw %es\n\t"
		"int $0x13\n\t"
		"pushfw\n\t"
		"cli\n\t"
	);

	in_stack.ax.h = 2; /* INT 13,2 - Read disk sector */
	in_stack.ax.l = 1; /* Read one sector */
	in_stack.cx.h = cylinder & 0xff;
	in_stack.cx.l = ( ( cylinder >> 8 ) & 0x3 ) | sector;
	in_stack.dx.h = head;
	in_stack.dx.l = drive;
	in_stack.buffer.segment = SEGMENT ( buf );
	in_stack.buffer.offset = OFFSET ( buf );
	ret_ax.word = real_call ( rm_pcbios_disk_read, &in_stack, &out_stack );
	return ( out_stack.flags.word & CF ) ? ret_ax.word : 0;
}
#endif /* TRY_FLOPPY_FIRST */

#endif /* PCBIOS */
